"""
Module containing functionality to support event-based microservices
"""

from .consumer import Consumer
from .consumer_exception import ConsumerException
from .event_handler import EventHandler

__all__ = ["Consumer", "ConsumerException", "EventHandler"]



"""
Custom wrapper for sqlalchemy
"""

import os
import importlib
from alembic import command
from alembic.config import Config
from alembic.util.exc import AutogenerateDiffsDetected
from sqlalchemy import create_engine, text, select
from sqlalchemy.orm import Session
from sqlalchemy.exc import NoResultFound
from loguru import logger
from .search_filters import SearchFilters

class SQLAdministrator:
    """
    sqlalchemy wrapper containing functionality for managing database connections
    and performing CRUD operations
    """

    def __init__(self):
        self.__connection = None
        self.__session = None

        self.__host = os.environ["DB_HOSTNAME"]
        self.__port = os.environ["DB_PORT"]
        self.__username = os.environ["DB_USER"]
        self.__password = os.environ["DB_PASSWORD"]
        self.__database = os.environ["DB_NAME"]

        self.__dsn = (
            f"mysql://{self.__username}:{self.__password}"
            f"@{self.__host}:{self.__port}/{self.__database}"
        )
        self.__engine = create_engine(self.__dsn)

        self.__alembic_root = os.environ["ALEMBIC_ROOT"]
        self.__table_metadata = getattr(importlib.import_module("model"), "Base").metadata

    def connect(self):
        """
        Establish a new database connection using the configured DSN if one has not
        already been established.
        """
        if not self.__connection:
            self.__connection = self.__engine.connect()
            self.__session = Session(self.__engine)

        return self.__connection

    def close(self):
        """
        Terminate the active database connection if one exists
        """
        if self.__connection:
            self.__session.close()
            self.__connection.close()

            self.__session = None
            self.__connection = None

    def commit(self):
        """
        Commit chznges made during the active session, if one exists
        """
        if self.__session:
            self.__session.commit()

    def rollback(self):
        """
        Rollback chznges made during the active session, if one exists
        """
        if self.__session:
            self.__session.rollback()

    def execute(self, query):
        """
        Execute an arbitrary SQL command over the open connection
        """
        if self.__connection:
            result = self.__connection.execute(text(query))
            return result.mappings().all()

        return None

    def create(self, resource):
        """
        Persist the specified resource to the database
        """
        if self.__session:
            self.__session.add(resource)
            self.__session.commit()

            return True

        return None

    def search(self, model, filters: SearchFilters):
        """
        Return a list of resources matching the specified search filters
        """
        query = select(model)
        for criteria in filters.criteria:
            if criteria["comparison"] == "==":
                query = query.where(criteria["field"] == criteria["value"])
            elif criteria["comparison"] == "!=":
                query = query.where(criteria["field"] != criteria["value"])
            elif criteria["comparison"] == "<":
                query = query.where(criteria["field"] < criteria["value"])
            elif criteria["comparison"] == "<=":
                query = query.where(criteria["field"] <= criteria["value"])
            elif criteria["comparison"] == ">":
                query = query.where(criteria["field"] > criteria["value"])
            elif criteria["comparison"] == ">=":
                query = query.where(criteria["field"] >= criteria["value"])

        resources = []
        for resource in self.__session.scalars(query):
            resources.append(resource)

        return resources

    def retrieve(self, model, id_number):
        """
        Retrieve the instance of the specified model identified by the provided id
        """
        try:
            resource = self.__session.scalars(
                select(model).where(model.id == id_number)
            ).one()

            return resource
        except NoResultFound:
            return None

    def delete(self, model, id_number):
        """
        Remove the instance of the specified model identified by the provided id
        from the database
        """
        resource = self.retrieve(model, id_number)
        if resource:
            self.__session.delete(resource)

        return resource

    def migrate(self):
        """
        Generate and apply a new database revision if the project's model has changed
        """

        # Initialise alembic configuration
        alembic_config = Config(os.environ["ALEMBIC_INI"])
        alembic_config.set_main_option("script_location", str(self.__alembic_root))
        alembic_config.set_main_option("sqlalchemy.url", self.__dsn)

        # Load data definitions
        alembic_config.attributes["target_metadata"] = self.__table_metadata

        # Generate a new revision
        try:
            command.check(alembic_config)
        except AutogenerateDiffsDetected as e:
            logger.info(f"Model change detected: {e}. Proceeding with migration.")
            command.revision(alembic_config, autogenerate=True)
            command.upgrade(alembic_config, "head")



"""
Empty Message Schema
"""
from .rpc_fields import rpc_fields

Empty = {
	"type": "object",
	"allOf": [
		{"$ref": "#/definitions/rpc_fields"}
	],
    "unevaluatedProperties": False,
    "definitions": {
        "rpc_fields": rpc_fields
    }
}



"""
Allow dynamic specification of search criteria for SQLAlchemy queries
"""
class SearchFilters:
    """
    The SearchFilters class contains a list of desired search criteria and a set of helper
    functions for manupulating them
    """

    def __init__(self):
        self.criteria = []

    def equal(self, field, value):
        """
        Add an equality test to the search filters
        """

        self.__add_criteria(field, "==",value)

    def at_least(self, field, value):
        """
        Add a greater than or equal test to the search filters
        """

        self.__add_criteria(field, ">=",value)

    def at_most(self, field, value):
        """
        Add a less than or equal test to the search filters
        """

        self.__add_criteria(field, "<=",value)

    def before(self, field, value):
        """
        Alias for at_most to be used for datetime fields
        """

        self.at_most(field, value)

    def after(self, field, value):
        """
        Alias for at_least to be used for datetime fields
        """

        self.at_least(field, value)

    def __add_criteria(self, field, comparison, value):
        self.criteria.append({
            "field": field,
            "comparison": comparison,
            "value": value
        })



"""
SQLAlchemy convenience wrappers for TickeTing microservices
"""

from .search_filters import SearchFilters
from .sql_administrator import SQLAdministrator

__all__ = ["SearchFilters", "SQLAdministrator"]



"""
consumer module
"""
from abc import ABC, abstractmethod

class Consumer(ABC):
    """
    The consumer class is an abstract class from which all event consumers must inherit
    """
    @abstractmethod
    def consume(self, channel, body):
        """
        Consumes a message from the queue and takes some action as a result
        """



"""
consumer_exception Module
"""

class ConsumerException(Exception):
    """
    A ConsumerException should be raised by any consumers if they encounter an error
    while processing a message received from the broker. This allows the EventHandler
    to decide whether to ACK or NACK the message and respond to RPC calls. 
    """
    def __init__(self, message):
        self.message = message
        super().__init__(self.message)



"""
event_handler Module
"""

import os
import time
import importlib
import json
import yaml
import pika
import jsonschema
from loguru import logger
from ticketing_service_core.orm import SQLAdministrator
from .consumer_exception import ConsumerException

class EventHandler:
    """
    The EventHandler sets up queue bindings and routes message handling to the
    relevant callback
    """

    def __init__(self, service: str):
        self.__rabbitmq = None
        self.__routing_map = {}
        self.__sql = SQLAdministrator()

        # Load events configuration
        with open(
            os.environ["EVENTS_CONFIG"], "r",
            encoding="utf-8"
        ) as events_file:
            config = yaml.safe_load(events_file)
            self.__events = config["events"]
            self.__message_queue = f"{service}.service.queue"


    def run(self):
        """
        Setup queue bindings and start message consumption
        """

        # Open a RabbitMQ connection if one does not exists
        while not self.__rabbitmq:
            logger.info("Attempting connection to RabbitMQ")
            try:
                self.__connect_to_rabbitmq()
                logger.info("Connection to RabbitMQ successful!")
            except pika.exceptions.AMQPConnectionError as e:
                logger.info(f"Failed to connect to RabbitMQ: {e}. Retrying.")
                time.sleep(1)

        channel = self.__rabbitmq.channel()

        # Ensure a consumer only receives a single message at a time
        channel.basic_qos(prefetch_count=1)

        # Setup Queue Bindings
        bindings = []
        for event in self.__events:
            # Initialise routing key
            action = (f".{event['action']}" if ("action" in event) else "")
            routing_key = f"{event['resource']}{action}"

            # If the service queue has already been bound to the target exchange
            # with this routing key
            if f"{event["exchange"]}:routing_key" in bindings:
                continue

            # Bind microservice queue to the indicated exchange
            logger.info(
                f"Binding {self.__message_queue} to {event['exchange']} exchange"
                f" with '{routing_key}' routing key"
            )

            try:
                channel.queue_bind(
                    exchange=event["exchange"],
                    queue=self.__message_queue,
                    routing_key=routing_key
                )

                bindings.append(f"{event["exchange"]}:routing_key")
            except pika.exceptions.ChannelClosedByBroker as e:
                logger.warning(
                    f"Could not bind service queue to {event["exchange"]} exchange: {e}"
                )

                channel = self.__rabbitmq.channel()

                # Ensure a consumer only receives a single message at a time
                channel.basic_qos(prefetch_count=1)


        # Start message consumption
        for event in self.__events:
            # Initialise routing key
            action = (f".{event['action']}" if ("action" in event) else "")
            routing_key = f"{event['resource']}{action}"

            # Add routing map entry
            if f"{event["exchange"]}:{routing_key}" not in self.__routing_map:
                self.__routing_map[f"{event["exchange"]}:{routing_key}"] = []

            # Dynamically import consumer module and instantiate consumer class
            consumer = getattr(
                importlib.import_module(os.environ["CONSUMER_MODULE"]),
                event["consumer"]
            )

            # Dynamically import excepted message schema
            request_schema = getattr(
                importlib.import_module("ticketing_service_core.schema"),
                event["request"]
            )

            self.__routing_map[f"{event["exchange"]}:{routing_key}"].append({
                "handler": consumer(),
                "request": request_schema
            })

        # Setup message consumption
        channel.basic_consume(
            queue=self.__message_queue,
            on_message_callback=self.__route_event,
            auto_ack=False
        )

        # Start consumption loop
        logger.info(
            f"Listening for nessages to {self.__message_queue}"
        )

        try:
            channel.start_consuming()
            channel.close()

            self.__rabbitmq = None
        except pika.exceptions.ConnectionClosedByBroker as e:
            logger.info("RabbitMQ went away. Attempting to reconnext")

            self.__rabbitmq = None
            self.run()


    def stop(self):
        """
        Close RabbitMQ connection, cleaning up queues if necessary
        """

        # Close the open RabbitMQ connection if one exists
        if self.__rabbitmq:
            self.__rabbitmq.close()

    def __connect_to_rabbitmq(self):
        # Setup blocking connection to RabbitMQ
        self.__rabbitmq = pika.BlockingConnection(
            pika.ConnectionParameters(
                host=os.environ["MQ_HOSTNAME"],
                port=os.environ["MQ_PORT"],
                credentials=pika.PlainCredentials(
                    os.environ["MQ_USERNAME"],
                    os.environ["MQ_PASSWORD"]
                )
            )
        )

    def __route_event(self, channel, method, properties, body):
        for consumer in self.__routing_map[f"{method.exchange}:{method.routing_key}"]:
            processing_error = None
            response = None
            message = {}

            try:
                # Validate message body
                message = json.loads(body)
                jsonschema.validate(instance=message, schema=consumer["request"])

                self.__sql.connect()

                # Process message
                response = consumer["handler"].consume(
                    channel=channel,
                    body=message,
                    sql=self.__sql
                )
            except (ValueError, jsonschema.ValidationError) as e:
                logger.warning(f"Parsing message body failed: {e}")
                response = {
                    "processing_error":
                    "The message body did not match the excepted schema."
                }
            except ConsumerException as e:
                processing_error = f"Comsumer threw an exception: {e}"
                response = {
                    "processing_error":
                    "The consumer encountered an error while processing this message."
                }

            # Send response to message when called in an RPC context
            if "reply_to" in message:
                if "correlation_id" in message:
                    # Add correlation id to response
                    response["correlation_id"] = message["correlation_id"]

                channel.basic_publish(
                    exchange="",
                    routing_key=message["reply_to"],
                    body=json.dumps(response)
                )

        # Handle any processing errors
        if processing_error:
            self.__sql.rollback()

            logger.error(
                f"Can't process message to '{method.exchange}' exchange "
                f"with routing key {method.routing_key}. "
                f"{processing_error}"
            )

            #Thottle redelivery to prevent flooding service
            time.sleep(1)

            # Send NACK to broker if message could not be processed
            channel.basic_nack(
                delivery_tag=method.delivery_tag,
                requeue=True
            )

            logger.info("NACKed message. Waiting for redelivery.")
        else:
            self.__sql.commit()

            # Send ACK to broker upon successful processing
            channel.basic_ack(delivery_tag=method.delivery_tag)

        self.__sql.close()



"""
Advertisement Schema
"""
Advertisement = {
	"type": "object",
	"properties": {
		"host": {"type": "string"},
		"event": {"type": "string"},
		"name": {"type": "string"},
		"description": {"type": "string"},
		"status": {"type": "string"}
	},
    "unevaluatedProperties": True
}
